#include "ntifs.h"
#include "test_driver.h"

typedef PVOID
(*PsGetCurrentProcessWow64Process_Fnc) (
    VOID
    );

typedef struct _DevExt
{
    PsGetCurrentProcessWow64Process_Fnc pPsGetCurrentProcessWow64Process;
}DevExt_type;

void DriverUnload(IN PDRIVER_OBJECT DriverObject)
{
}

NTSTATUS CompleteIrp(PIRP Irp, NTSTATUS status, ULONG information)
{
    Irp->IoStatus.Information = information;
    Irp->IoStatus.Status = status;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return status;
}

NTSTATUS DoTest_i386(IN PDEVICE_OBJECT  DeviceObject,
                     IN PVOID pBuffer, 
                     IN ULONG sizeIn, 
                     IN ULONG maxBufferSize, 
                     OUT ULONG * pSizeOut)
{
    // user is under WOW
    if (sizeIn < sizeof(OldRequest_i386_type))
    return STATUS_BUFFER_OVERFLOW;

    if (maxBufferSize < sizeof(OldRequest_i386_type))
    return STATUS_BUFFER_OVERFLOW;

    {
        OldRequest_i386_type * p = (OldRequest_i386_type *)pBuffer;
        p->c = p->a + p->b;
    }
    *pSizeOut = sizeof(OldRequest_i386_type);
    return STATUS_SUCCESS;
}

NTSTATUS DoTest(IN PDEVICE_OBJECT  DeviceObject,
                IN PVOID pBuffer, 
                IN ULONG sizeIn, 
                IN ULONG maxBufferSize, 
                OUT ULONG * pSizeOut)
{
    DevExt_type * pDevExt=
        (DevExt_type*)DeviceObject->DeviceExtension;

    *pSizeOut = 0;
#if defined (_M_AMD64)
    // only for AMD64 builds
    if (pDevExt->pPsGetCurrentProcessWow64Process())
    {
        return DoTest_i386();
    }

    // user is true amd64 guy
    if (sizeIn < sizeof(OldRequest_amd64_type))
        return STATUS_BUFFER_OVERFLOW;

    if (maxBufferSize < sizeof(OldRequest_amd64_type))
        return STATUS_BUFFER_OVERFLOW;

    {
        OldRequest_amd64_type * p = (OldRequest_amd64_type *)pBuffer;
        p->c = p->a + p->b;
    }
    *pSizeOut = sizeof(OldRequest_amd64_type);
    return STATUS_SUCCESS;
#else
    return DoTest_i386(DeviceObject,
        pBuffer, 
        sizeIn, 
        maxBufferSize, 
        pSizeOut);
#endif
}

NTSTATUS
DispatchControl(IN  PDEVICE_OBJECT  DeviceObject,
                IN  PIRP            Irp)
{
    PIO_STACK_LOCATION irpSp = IoGetCurrentIrpStackLocation(Irp);

    NTSTATUS status = 0;
    ULONG code = 0;
    ULONG outBufSize = irpSp->Parameters.DeviceIoControl.OutputBufferLength;
    ULONG inBufSize = irpSp->Parameters.DeviceIoControl.InputBufferLength;

    ULONG outSizeRes = 0;
    PVOID pBuffer = Irp->AssociatedIrp.SystemBuffer;

    code = irpSp->Parameters.DeviceIoControl.IoControlCode;

    status = STATUS_INVALID_DEVICE_REQUEST;
    switch (code)
    {
        case TESTWOW_TEST:
            {
                status = DoTest(DeviceObject, 
                                pBuffer, 
                                inBufSize, 
                                outBufSize, 
                                &outSizeRes);
                break;
            }
    }
    return CompleteIrp(Irp, status, outSizeRes); 
}

NTSTATUS
DispatchFnc(IN  PDEVICE_OBJECT  DeviceObject,
            IN  PIRP            Irp)
{
    PIO_STACK_LOCATION irpSp = IoGetCurrentIrpStackLocation(Irp);
    switch(irpSp->MajorFunction)
    {
    case IRP_MJ_CREATE:
        return CompleteIrp(Irp, STATUS_SUCCESS, 0);
    case IRP_MJ_CLOSE:
        return CompleteIrp(Irp, STATUS_SUCCESS, 0);
    case IRP_MJ_DEVICE_CONTROL:
        return DispatchControl(DeviceObject, Irp);
    default:;
    }
    return CompleteIrp(Irp, STATUS_INVALID_DEVICE_REQUEST, 0);
}

NTSTATUS DriverEntry(IN  PDRIVER_OBJECT  DriverObject,
                    IN  PUNICODE_STRING RegistryPath)
{
    UNICODE_STRING deviceName, symlink, psGetCurrentProcessWow64ProcessName;
    NTSTATUS status = 0;
    DevExt_type * pDevExt = 0;
    PDEVICE_OBJECT deviceObject = 0;
    int i =0;

    // set unload routine
    DriverObject->DriverUnload  = DriverUnload;
    for (i=0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++)
    {
        DriverObject->MajorFunction[i] = DispatchFnc;
    }

    RtlInitUnicodeString(&deviceName, L"\\Device\\TestWOWDriverDevice");
    // create device
    status = IoCreateDevice(
        DriverObject,
        sizeof(DevExt_type),
        &deviceName,
        FILE_DEVICE_UNKNOWN,
        0,
        FALSE,
        &deviceObject
        );

    if (!NT_SUCCESS(status))
        return status;

    // init device
    pDevExt = (DevExt_type*)deviceObject->DeviceExtension;

    // get routine
    RtlInitUnicodeString(&psGetCurrentProcessWow64ProcessName, L"PsGetCurrentProcessWow64Process");
    pDevExt->pPsGetCurrentProcessWow64Process = MmGetSystemRoutineAddress(&psGetCurrentProcessWow64ProcessName);

#if defined (_M_AMD64)
    if (!pDevExt->pPsGetCurrentProcessWow64Process)
    {
        status = STATUS_PROCEDURE_NOT_FOUND;
        goto clear;
    }
#endif
    SetFlag(deviceObject->Flags, DO_BUFFERED_IO);
    ClearFlag(deviceObject->Flags, DO_DEVICE_INITIALIZING);

    // create symlink
    RtlInitUnicodeString(&symlink, TESTWOW_TEST_SYMBOLIC_LINK);
    status = IoCreateSymbolicLink(&symlink, &deviceName);
    if (!NT_SUCCESS(status))
        goto clear;
        

clear:

    // cleanup
    if (!NT_SUCCESS(status))
    {
        IoDeleteDevice(deviceObject);
    }
    return status;
}